@@ -85,6 +85,7 @@ impl SearchScope {
         SearchScope { entries }
     }
 
+    /// Build a search scope spanning the entire crate graph of files.
     fn crate_graph(db: &RootDatabase) -> SearchScope {
         let mut entries = FxHashMap::default();
 
@@ -98,6 +99,7 @@ impl SearchScope {
         SearchScope { entries }
     }
 
+    /// Build a search scope spanning all the reverse dependencies of the given crate.
     fn reverse_dependencies(db: &RootDatabase, of: hir::Crate) -> SearchScope {
         let mut entries = FxHashMap::default();
         for rev_dep in of.transitive_reverse_dependencies(db) {
@@ -109,6 +111,7 @@ impl SearchScope {
         SearchScope { entries }
     }
 
+    /// Build a search scope spanning the given crate.
     fn krate(db: &RootDatabase, of: hir::Crate) -> SearchScope {
         let root_file = of.root_file(db);
         let source_root_id = db.file_source_root(root_file);
@@ -118,55 +121,55 @@ impl SearchScope {
         }
     }
 
-    fn module(db: &RootDatabase, module: hir::Module) -> SearchScope {
+    /// Build a search scope spanning the given module and all its submodules.
+    fn module_and_children(db: &RootDatabase, module: hir::Module) -> SearchScope {
         let mut entries = FxHashMap::default();
 
-        let mut to_visit = vec![module];
-        let mut is_first = true;
+        let (file_id, range) = {
+            let InFile { file_id, value } = module.definition_source(db);
+            if let Some((file_id, call_source)) = file_id.original_call_node(db) {
+                (file_id, Some(call_source.text_range()))
+            } else {
+                (
+                    file_id.original_file(db),
+                    match value {
+                        ModuleSource::SourceFile(_) => None,
+                        ModuleSource::Module(it) => Some(it.syntax().text_range()),
+                        ModuleSource::BlockExpr(it) => Some(it.syntax().text_range()),
+                    },
+                )
+            }
+        };
+        entries.insert(file_id, range);
+
+        let mut to_visit: Vec<_> = module.children(db).collect();
         while let Some(module) = to_visit.pop() {
-            let src = module.definition_source(db);
-            let file_id = src.file_id.original_file(db);
-            match src.value {
-                ModuleSource::Module(m) => {
-                    if is_first {
-                        let range = Some(m.syntax().text_range());
-                        entries.insert(file_id, range);
-                    } else {
-                        // We have already added the enclosing file to the search scope,
-                        // so do nothing.
-                    }
-                }
-                ModuleSource::BlockExpr(b) => {
-                    if is_first {
-                        let range = Some(b.syntax().text_range());
-                        entries.insert(file_id, range);
-                    } else {
-                        // We have already added the enclosing file to the search scope,
-                        // so do nothing.
-                    }
-                }
-                ModuleSource::SourceFile(_) => {
-                    entries.insert(file_id, None);
-                }
-            };
-            is_first = false;
+            if let InFile { file_id, value: ModuleSource::SourceFile(_) } =
+                module.definition_source(db)
+            {
+                entries.insert(file_id.original_file(db), None);
+            }
             to_visit.extend(module.children(db));
         }
         SearchScope { entries }
     }
 
+    /// Build an empty search scope.
     pub fn empty() -> SearchScope {
         SearchScope::new(FxHashMap::default())
     }
 
+    /// Build a empty search scope spanning the given file.
     pub fn single_file(file: FileId) -> SearchScope {
         SearchScope::new(std::iter::once((file, None)).collect())
     }
 
+    /// Build a empty search scope spanning the text range of the given file.
     pub fn file_range(range: FileRange) -> SearchScope {
         SearchScope::new(std::iter::once((range.file_id, Some(range.range))).collect())
     }
 
+    /// Build a empty search scope spanning the given files.
     pub fn files(files: &[FileId]) -> SearchScope {
         SearchScope::new(files.iter().map(|f| (*f, None)).collect())
     }
@@ -177,29 +180,23 @@ impl SearchScope {
             mem::swap(&mut small, &mut large)
         }
 
+        let intersect_ranges =
+            |r1: Option<TextRange>, r2: Option<TextRange>| -> Option<Option<TextRange>> {
+                match (r1, r2) {
+                    (None, r) | (r, None) => Some(r),
+                    (Some(r1), Some(r2)) => r1.intersect(r2).map(Some),
+                }
+            };
         let res = small
             .iter()
-            .filter_map(|(file_id, r1)| {
-                let r2 = large.get(file_id)?;
-                let r = intersect_ranges(*r1, *r2)?;
-                Some((*file_id, r))
+            .filter_map(|(&file_id, &r1)| {
+                let &r2 = large.get(&file_id)?;
+                let r = intersect_ranges(r1, r2)?;
+                Some((file_id, r))
             })
             .collect();
 
-        return SearchScope::new(res);
-
-        fn intersect_ranges(
-            r1: Option<TextRange>,
-            r2: Option<TextRange>,
-        ) -> Option<Option<TextRange>> {
-            match (r1, r2) {
-                (None, r) | (r, None) => Some(r),
-                (Some(r1), Some(r2)) => {
-                    let r = r1.intersect(r2)?;
-                    Some(Some(r))
-                }
-            }
-        }
+        SearchScope::new(res)
     }
 }
 
@@ -223,7 +220,7 @@ impl Definition {
         // def is crate root
         // FIXME: We don't do searches for crates currently, as a crate does not actually have a single name
         if let &Definition::Module(module) = self {
-            if module.crate_root(db) == module {
+            if module.is_crate_root(db) {
                 return SearchScope::reverse_dependencies(db, module.krate());
             }
         }
@@ -271,7 +268,7 @@ impl Definition {
         }
 
         if let Definition::Macro(macro_def) = self {
-            return match macro_def.kind() {
+            return match macro_def.kind(db) {
                 hir::MacroKind::Declarative => {
                     if macro_def.attrs(db).by_key("macro_export").exists() {
                         SearchScope::reverse_dependencies(db, module.krate())
@@ -282,7 +279,8 @@ impl Definition {
                 hir::MacroKind::BuiltIn => SearchScope::crate_graph(db),
                 // FIXME: We don't actually see derives in derive attributes as these do not
                 // expand to something that references the derive macro in the output.
-                // We could get around this by emitting dummy `use DeriveMacroPathHere as _;` items maybe?
+                // We could get around this by doing pseudo expansions for proc_macro_derive like we
+                // do for the derive attribute
                 hir::MacroKind::Derive | hir::MacroKind::Attr | hir::MacroKind::ProcMacro => {
                     SearchScope::reverse_dependencies(db, module.krate())
                 }
@@ -294,7 +292,7 @@ impl Definition {
             return SearchScope::reverse_dependencies(db, module.krate());
         }
         if let Some(Visibility::Module(module)) = vis {
-            return SearchScope::module(db, module.into());
+            return SearchScope::module_and_children(db, module.into());
         }
 
         let range = match module_source {
@@ -310,6 +308,10 @@ impl Definition {
 
     pub fn usages<'a>(self, sema: &'a Semantics<RootDatabase>) -> FindUsages<'a> {
         FindUsages {
+            local_repr: match self {
+                Definition::Local(local) => Some(local.representative(sema.db)),
+                _ => None,
+            },
             def: self,
             sema,
             scope: None,
@@ -325,6 +327,7 @@ pub struct FindUsages<'a> {
     sema: &'a Semantics<'a, RootDatabase>,
     scope: Option<SearchScope>,
     include_self_kw_refs: Option<hir::Type>,
+    local_repr: Option<hir::Local>,
     search_self_mod: bool,
 }
 
@@ -336,10 +339,12 @@ impl<'a> FindUsages<'a> {
         self
     }
 
+    /// Limit the search to a given [`SearchScope`].
     pub fn in_scope(self, scope: SearchScope) -> FindUsages<'a> {
         self.set_scope(Some(scope))
     }
 
+    /// Limit the search to a given [`SearchScope`].
     pub fn set_scope(mut self, scope: Option<SearchScope>) -> FindUsages<'a> {
         assert!(self.scope.is_none());
         self.scope = scope;
@@ -378,7 +383,7 @@ impl<'a> FindUsages<'a> {
 
         let name = match self.def {
             // special case crate modules as these do not have a proper name
-            Definition::Module(module) if module.crate_root(self.sema.db) == module => {
+            Definition::Module(module) if module.is_crate_root(self.sema.db) => {
                 // FIXME: This assumes the crate name is always equal to its display name when it really isn't
                 module
                     .krate()
@@ -415,6 +420,7 @@ impl<'a> FindUsages<'a> {
                 Some(offset)
             })
         }
+
         fn scope_files<'a>(
             sema: &'a Semantics<RootDatabase>,
             scope: &'a SearchScope,
@@ -428,6 +434,12 @@ impl<'a> FindUsages<'a> {
             })
         }
 
+        // FIXME: There should be optimization potential here
+        // Currently we try to descend everything we find which
+        // means we call `Semantics::descend_into_macros` on
+        // every textual hit. That function is notoriously
+        // expensive even for things that do not get down mapped
+        // into macros.
         for (text, file_id, search_range) in scope_files(sema, &search_scope) {
             let tree = Lazy::new(move || sema.parse(file_id).syntax().clone());
 
@@ -458,9 +470,10 @@ impl<'a> FindUsages<'a> {
         // Search for `super` and `crate` resolving to our module
         match self.def {
             Definition::Module(module) => {
-                let scope = search_scope.intersection(&SearchScope::module(self.sema.db, module));
+                let scope = search_scope
+                    .intersection(&SearchScope::module_and_children(self.sema.db, module));
 
-                let is_crate_root = module.crate_root(self.sema.db) == module;
+                let is_crate_root = module.is_crate_root(self.sema.db);
 
                 for (text, file_id, search_range) in scope_files(sema, &scope) {
                     let tree = Lazy::new(move || sema.parse(file_id).syntax().clone());
@@ -593,6 +606,19 @@ impl<'a> FindUsages<'a> {
         sink: &mut dyn FnMut(FileId, FileReference) -> bool,
     ) -> bool {
         match NameRefClass::classify(self.sema, name_ref) {
+            Some(NameRefClass::Definition(def @ Definition::Local(local)))
+                if matches!(
+                    self.local_repr, Some(repr) if repr == local.representative(self.sema.db)
+                ) =>
+            {
+                let FileRange { file_id, range } = self.sema.original_range(name_ref.syntax());
+                let reference = FileReference {
+                    range,
+                    name: ast::NameLike::NameRef(name_ref.clone()),
+                    category: ReferenceCategory::new(&def, name_ref),
+                };
+                sink(file_id, reference)
+            }
             Some(NameRefClass::Definition(def)) if def == self.def => {
                 let FileRange { file_id, range } = self.sema.original_range(name_ref.syntax());
                 let reference = FileReference {
@@ -622,7 +648,7 @@ impl<'a> FindUsages<'a> {
                     Definition::Field(_) if field == self.def => {
                         ReferenceCategory::new(&field, name_ref)
                     }
-                    Definition::Local(l) if local == l => {
+                    Definition::Local(_) if matches!(self.local_repr, Some(repr) if repr == local.representative(self.sema.db)) => {
                         ReferenceCategory::new(&Definition::Local(local), name_ref)
                     }
                     _ => return false,
@@ -667,6 +693,21 @@ impl<'a> FindUsages<'a> {
                 };
                 sink(file_id, reference)
             }
+            Some(NameClass::Definition(def @ Definition::Local(local))) if def != self.def => {
+                if matches!(
+                    self.local_repr,
+                    Some(repr) if local.representative(self.sema.db) == repr
+                ) {
+                    let FileRange { file_id, range } = self.sema.original_range(name.syntax());
+                    let reference = FileReference {
+                        range,
+                        name: ast::NameLike::Name(name.clone()),
+                        category: None,
+                    };
+                    return sink(file_id, reference);
+                }
+                false
+            }
             // Resolve trait impl function definitions to the trait definition's version if self.def is the trait definition's
             Some(NameClass::Definition(def)) if def != self.def => {
                 /* poor man's try block */
@@ -701,13 +742,7 @@ fn def_to_ty(sema: &Semantics<RootDatabase>, def: &Definition) -> Option<hir::Ty
     match def {
         Definition::Adt(adt) => Some(adt.ty(sema.db)),
         Definition::TypeAlias(it) => Some(it.ty(sema.db)),
-        Definition::BuiltinType(it) => {
-            let graph = sema.db.crate_graph();
-            let krate = graph.iter().next()?;
-            let root_file = graph[krate].root_file_id;
-            let module = sema.to_module_def(root_file)?;
-            Some(it.ty(sema.db, module))
-        }
+        Definition::BuiltinType(it) => Some(it.ty(sema.db)),
         Definition::SelfType(it) => Some(it.self_ty(sema.db)),
         _ => None,
     }
